from collections import UserDict

import httpx
from yarl import URL

API_URL = URL("https://api.urbandictionary.com/v0/uncacheable")


class CustomDict(UserDict):
    def __getattr__(self, attr):
        if attr in self.data:
            return self.data.get(attr)


def _sync_get_votes(ids: list):
    url = API_URL.with_query(ids=",".join(map(str, ids)))
    res = httpx.get(url.human_repr())
    return [CustomDict(i) for i in res.json().get('thumbs')]

async def _async_get_votes(ids: list):
    url = API_URL.with_query(ids=",".join(map(str, ids)))
    async with httpx.AsyncClient() as client:
        res = await client.get(url.human_repr())
        return [CustomDict(i) for i in res.json().get('thumbs')]

def _search_defid(id: int, votes: list[CustomDict]):
    for d in votes:
        if int(d.defid) == id:
            return d
    raise KeyError


MONTHS = {
    "January": 1,
    "February": 2,
    "March": 3,
    "April": 4,
    "May": 5,
    "June": 6,
    "July": 7,
    "August": 8,
    "September": 9,
    "October": 10,
    "November": 11,
    "December": 12,
}


from dataclasses import dataclass, asdict
from datetime import datetime


@dataclass
class Author:
    name: str
    url: str


@dataclass(init=True)
class Definition:
    defid: int
    name: str
    meaning: str
    example: str
    author: Author
    date: datetime
    upvote: int = None
    downvote: int = None

    def as_dict(self):
        """Converts the dataclass to a `dict` object and returns it."""
        return asdict(self)



from setuptools import setup, find_packages


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
    name="urban_scraper",
    version="1.2",
    packages=find_packages(include=["urban_scraper", "urban_scraper.*"]),
    install_requires=["aiofiles", "aiopath", "pickledb", "selectolax", "yarl", "httpx"],
    author="Myxi",
    author_email="myxi@duck.com",
    url="https://github.com/m-y-x-i/urban-scraper",
    description="Easy to use Urban Dictionary asynchronous and synchronous scraper.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.9",
    license="GNU GPLv3",
)



import json
import time
from datetime import datetime
from typing import AsyncIterator, Awaitable

import aiofiles
import aiopath
import httpx
from selectolax.parser import HTMLParser
from yarl import URL

from urban_scraper._dataclasses import Author, Definition
from urban_scraper._months import MONTHS

from ._get_votes import _async_get_votes, _search_defid


def anext(async_iterator: AsyncIterator) -> Awaitable:
    return async_iterator.__anext__()


async def define(query: str) -> AsyncIterator[Definition]:
    """Returns an async iterator of all the `Definition` it could fetch."""

    gen = await Define.construct(query)
    return gen


class Define:
    @classmethod
    async def construct(cls, query: str):
        self = Define()
        self.query = query.strip()
        self.url = URL("https://www.urbandictionary.com/define.php")
        self.query_url = self.url.with_query(term=self.query)
        await self._ensure()
        return self

    def __aiter__(self):
        return self

    async def __anext__(self):
        return await anext(self.iterator)

    async def _ensure(self):
        path = aiopath.AsyncPath(__file__).parent.joinpath('_cache')
        if not await path.exists():
            await path.mkdir()
        self.path = path.joinpath("cache.db")
        if not await self.path.exists():
            async with aiofiles.open(self.path, "w") as cache_file:
                await cache_file.write("{}")
                await cache_file.close()
        self.iterator = self._iterdefs()

    async def _get_cache(self, key: str):
        async with aiofiles.open(self.path, "r") as cache_file:
            obj = json.loads(await cache_file.read())
            self.obj = obj
            if key.lower() in obj:
                return obj.get(key.lower())
            else:
                return None

    async def _get_html(self):
        html = await self._get_cache(self.query.lower())
        if html:
            if round(time.time() - html.get("time")) < 3600:
                return html.get("data")
        else:
            async with aiofiles.open(self.path, "w") as cache_file:
                html = await self._fetch_html()
                cache_time = time.time()
                self.obj[self.query.lower()] = dict(data=html, time=cache_time)
                await cache_file.write(json.dumps(self.obj))
                await cache_file.close()
            return html

    async def _fetch_html(self):
        async with httpx.AsyncClient() as client:
            res = await client.get(str(self.query_url))
            return res.text

    async def _iterdefs(self):
        html = await self._get_html()
        parser = HTMLParser(html, True)
        nodes = parser.css(".def-panel")
        defids = [int(i.attrs.get("data-defid")) for i in nodes]
        votes = await _async_get_votes(defids)
        for index, node in enumerate(parser.css(".def-panel")):
            contributor = node.css_first(".contributor")
            author = contributor.css_first("a")
            date = [i.strip(",") for i in contributor.text().split()[-3:]]

            defid = defids[index]
            name = node.css_first(".word").text()
            meaning = node.css_first(".meaning").text()
            example = node.css_first(".example").text()
            author_name = author.text()
            author_url = str(
                URL("https://www.urbandictionary.com").join(URL(author.attrs["href"]))
            )
            _datetime = datetime(
                year=int(date[2]), month=MONTHS.get(date[0]), day=int(date[1])
            )
            upvote = _search_defid(defid, votes).up
            downvote = _search_defid(defid, votes).down
            yield Definition(
                defid=defid,
                name=name,
                meaning=meaning,
                example=example,
                author=Author(name=author_name, url=author_url),
                date=_datetime,
                upvote=upvote,
                downvote=downvote,
            )



import pathlib
import time
from datetime import datetime
from typing import Iterator

import httpx
import pickledb
from selectolax.parser import HTMLParser
from yarl import URL

from ._dataclasses import Author, Definition
from ._get_votes import _search_defid, _sync_get_votes
from ._months import MONTHS


def define(query: str) -> Iterator[Definition]:
    """Returns an iterator of all the `Definition` it could fetch."""
    gen = Define(query)
    return gen

class Caching:
    def __init__(self) -> None:
        self._ensure()

    def _ensure(self):
        path = pathlib.Path(__file__).parent.joinpath('_cache')
        if not path.exists():
            path.mkdir()
        self.pickle = pickledb.load(str(path.joinpath("cache.db")), True)

    def _get_html_cache(self):
        html = self.pickle.get(self.query.lower())
        if html is not False:
            if round(time.time() - html.get("time")) < 3600:
                return html.get("data")
            else:
                return None


class Define(Caching):
    def __init__(self, query: str):
        self.query = query.strip()
        self.url = URL("https://www.urbandictionary.com/define.php")
        self.query_url = self.url.with_query(term=self.query)
        super().__init__()
        self.iterator = self._iterdefs()

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.iterator)

    def _fetch_html(self):
        res = httpx.get(str(self.query_url))
        return res.text

    def _get_html(self):
        cache = self._get_html_cache()
        if cache:
            return cache
        else:
            html = self._fetch_html()
            cache_time = time.time()
            obj = dict(data=html, time=cache_time)
            self.pickle.set(self.query.lower(), obj)
            return html

    def _iterdefs(self):
        html = self._get_html()
        parser = HTMLParser(html, True)
        nodes = parser.css(".def-panel")
        defids = [int(i.attrs.get('data-defid')) for i in nodes]
        votes = _sync_get_votes(defids)
        for index, node in enumerate(nodes):
            contributor = node.css_first(".contributor")
            author = contributor.css_first("a")
            date = [i.strip(",") for i in contributor.text().split()[-3:]]

            defid = defids[index]
            name = node.css_first(".word").text()
            meaning = node.css_first(".meaning").text()
            example = node.css_first(".example").text()
            author_name = author.text()
            author_url = str(
                URL("https://www.urbandictionary.com").join(URL(author.attrs["href"]))
            )
            _datetime = datetime(
                year=int(date[2]), month=MONTHS.get(date[0]), day=int(date[1])
            )
            upvote = _search_defid(defid, votes).up
            downvote = _search_defid(defid, votes).down
            yield Definition(
                defid=defid,
                name=name,
                meaning=meaning,
                example=example,
                author=Author(name=author_name, url=author_url),
                date=_datetime,
                upvote=upvote,
                downvote=downvote
            )
